/*
 * VeriSilicon VPE GC620 process interface
 * Copyright (C) 2020 VeriSilicon Holdings Co., Ltd.
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "libavutil/avstring.h"
#include "libavutil/imgutils.h"
#include "libavutil/opt.h"
#include "libavutil/parseutils.h"
#include "libavutil/pixdesc.h"
#include "libavutil/pixfmt.h"
#include "libswscale/swscale_internal.h"

#include "avfilter.h"
#include "drawutils.h"
#include "formats.h"
#include "internal.h"
#include "framesync.h"
#include "video.h"

#include <libavutil/hwcontext.h>
#include <libavutil/hwcontext_vpe.h>
#include "vpi_types.h"
#include "vpi_api.h"
#include "log_ama.h"

#define OUT_MAX 7680
enum { R, G, B, A };
enum ANGLE { ROTATE_0_DEGREE = 0, ROTATE_90_DEGREE = 90, ROTATE_180_DEGREE = 180 , ROTATE_270_DEGREE = 270 };
typedef struct VpeGC620Context {
    const AVClass *class;
    const AVPixFmtDescriptor *desc;
    unsigned x, y;
    int nb_inputs;
    FFFrameSync fs;
    int p_type;
    char *p_type_str;
    int core_id;

    /* for stack */
    AVFrame **frames;
    char *layout;
    int shortest;

    /* for drawbox */
    int thickness;
    char *color_str;
    uint8_t yuv_color[4];
    uint8_t rgb_color[4];

    /* for overlay */
    int overlay_x;
    int overlay_y;

    /* for rotate */
    int angle;

    /* for csc */
    char *csc;

    VpiProcInitParams vpe_gc620_params;
    VpiProcHandle vpe_gc620_handle;

    AVBufferRef *hw_frame_ctx;
    AVBufferRef *hwdevice;

    int out_chan_index;

    char instance_name[64];
    /* for output resolution */
    int width;
    int height;
} VpeGC620Context;

static int vpe_init_hwdev(AVFilterContext *ctx)
{
    VpeGC620Context *s = ctx->priv;

    if (!s->hwdevice) {
        if (ctx->hw_device_ctx) {
            s->hwdevice = av_buffer_ref(ctx->hw_device_ctx);
            if (!s->hwdevice) {
                return AVERROR(ENOMEM);
            }
        } else {
            return AVERROR_UNKNOWN;
        }
    }

    return 0;
}

static int vpe_check_constrains(AVFilterContext *ctx)
{
    VpeGC620Context *s = ctx->priv;
    AVHWFramesConstraints *constrains;
    AVFilterLink *inlink    = ctx->inputs[0];
    AVVpeHWConfig *hwconfig = NULL;
    int ret = 0;

    hwconfig = av_hwdevice_hwconfig_alloc(s->hwdevice);
    if (!hwconfig) {
        ret = AVERROR(ENOMEM);
        goto error;
    }

    hwconfig->id = VPE_HW_TILE;

    constrains = av_hwdevice_get_hwframe_constraints(s->hwdevice, hwconfig);
    if (!constrains) {
        ret = AVERROR(ENOMEM);
        goto error;
    }

    if (inlink->w < constrains->min_width ||
        inlink->h < constrains->min_height ||
        inlink->w > constrains->max_width ||
        inlink->h > constrains->max_height) {
        av_log(ctx, AV_LOG_ERROR,
               "size %dx%d is not supported. "
               "constrains: width %d-%d height %d-%d.\n",
               inlink->w, inlink->h, constrains->min_width,
               constrains->max_width, constrains->min_height,
               constrains->max_height);
        ret = AVERROR(EINVAL);
        goto error;
    }

error:
    if (constrains)
        av_hwframe_constraints_free(&constrains);
    if (hwconfig)
        av_freep(&hwconfig);

    return ret;
}

static int vpe_init_hwctx(AVFilterContext *ctx)
{
    VpeGC620Context *s        = ctx->priv;
    VpiProcInitParams *param = &s->vpe_gc620_params;
    AVHWFramesContext *hw_frame_ctx;
    AVHWDeviceContext *device_ctx;
    AVVpeDeviceContext *device_hwctx;
    AVVpeFramesContext *frame_hwctx;
    AVFilterLink *inlink = ctx->inputs[0];
    int ret;

    /* hw device & frame init */
    ret = vpe_init_hwdev(ctx);
    if (ret < 0) {
        return ret;
    }

    if (inlink->hw_frames_ctx) {
        av_log(ctx, AV_LOG_DEBUG, "%s(%d)\n", __FUNCTION__, __LINE__);
        s->hw_frame_ctx = av_buffer_ref(inlink->hw_frames_ctx);
        if (!s->hw_frame_ctx) {
            ret = AVERROR(ENOMEM);
            return ret;
        }

        hw_frame_ctx = (AVHWFramesContext *)s->hw_frame_ctx->data;
    } else {
        av_log(ctx, AV_LOG_ERROR, "%s(%d)\n", __FUNCTION__, __LINE__);
        ret = AVERROR_UNKNOWN;
        return ret;
    }

    ret = vpe_check_constrains(ctx);
    if (ret < 0) {
        return ret;
    }

    device_ctx   = hw_frame_ctx->device_ctx;
    device_hwctx = device_ctx->hwctx;
    frame_hwctx  = hw_frame_ctx->hwctx;

    vpi_device_set_ctx(param, device_hwctx->vpe_dev_ctx, VPE_HW_DRAWBOX);
    param->mem_ctx = frame_hwctx->vpe_mem_ctx;
    av_log(ctx, AV_LOG_DEBUG, "%s: vpe_mem_ctx = %p.\n", __func__,
           frame_hwctx->vpe_mem_ctx);

    return 0;
}

static int vpe_2d_query_formats(AVFilterContext *ctx)
{
    VpeGC620Context *s                  = ctx->priv;
    AVVpeHWConfig *hwconfig            = NULL;
    AVHWFramesConstraints *constraints = NULL;
    AVFilterFormats *in_fmts;
    AVFilterFormats *out_fmts;
    int ret;
    int i;

    ret = vpe_init_hwdev(ctx);
    if (ret < 0) {
        goto error;
    }

    hwconfig = av_hwdevice_hwconfig_alloc(s->hwdevice);
    if (!hwconfig) {
        ret = AVERROR(ENOMEM);
        goto error;
    }

    hwconfig->id = VPE_HW_TILE;
    av_log(ctx, AV_LOG_DEBUG, "hwconfig->id = %d\n", hwconfig->id);

    constraints = av_hwdevice_get_hwframe_constraints(s->hwdevice, hwconfig);
    if (!constraints) {
        av_log(ctx, AV_LOG_ERROR, "failed to get hwframe constrains.\n");
        ret = AVERROR(ENOMEM);
        goto error;
    }

    av_log(ctx, AV_LOG_DEBUG, "%s(%d)\n", __FUNCTION__, __LINE__);

    out_fmts = ff_make_format_list(constraints->valid_hw_formats);
    ret      = ff_formats_ref(out_fmts, &ctx->outputs[0]->incfg.formats);
    av_log(ctx, AV_LOG_DEBUG, "%s(%d) ret = %d\n", __FUNCTION__, __LINE__, ret);
    if (ret < 0)
        goto error;

    in_fmts = ff_make_format_list(constraints->valid_hw_formats);
    i       = 0;
    while (constraints->valid_sw_formats[i] != AV_PIX_FMT_NONE) {
        ret = ff_add_format(&in_fmts, constraints->valid_sw_formats[i]);
        av_log(ctx, AV_LOG_DEBUG, "%s(%d) ret = %d\n", __FUNCTION__, __LINE__,
               ret);
        if (ret < 0)
            goto error;
        i++;
    }

    ret = ff_formats_ref(in_fmts, &ctx->inputs[0]->outcfg.formats);
    av_log(ctx, AV_LOG_DEBUG, "%s(%d) ret = %d\n", __FUNCTION__, __LINE__, ret);
    if (ret < 0)
        goto error;

    av_hwframe_constraints_free(&constraints);
    av_freep(&hwconfig);

    return 0;

error:
    av_log(ctx, AV_LOG_ERROR, "%s failed\n", __FUNCTION__);

    if (constraints)
        av_hwframe_constraints_free(&constraints);
    if (hwconfig)
        av_freep(&hwconfig);

    return ret;
}

static int query_formats(AVFilterContext *ctx)
{
    int ret;
    VpeGC620Context *s = ctx->priv;

    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GC, s->instance_name);
    ret = vpe_2d_query_formats(ctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();

    return ret;
}

static int get_vpifmt_from_avpixfmt(enum AVPixelFormat format)
{
    switch (format) {
        case AV_PIX_FMT_YUV420P:     return VPI_FORMAT_YUV420P;
        case AV_PIX_FMT_NV12:        return VPI_FORMAT_NV12;
        case AV_PIX_FMT_RGBP:        return VPI_FORMAT_RGB24_P;
        case AV_PIX_FMT_BGRA:        return VPI_FORMAT_BGRA;
        case AV_PIX_FMT_RGBA:        return VPI_FORMAT_RGBA;

        default:
            return VPI_FORMAT_NONE;
    }
}

static void set_instance_name(VpeGC620Context *s, const char* ffmpeg_name)
{
    char *ptr = strrchr(ffmpeg_name, '_');
    snprintf(s->instance_name, sizeof(s->instance_name), (ptr) ? "2d_ama%s" : "2d_ama@%p", (ptr) ? ptr : s);
}

static av_cold int vpe_2d_init(AVFilterContext *ctx)
{
    VpeGC620Context *s = ctx->priv;
    int i, ret;

    if (!strcmp(s->p_type_str, "none"))
        return AVERROR(EINVAL);

    if (!strcmp(s->p_type_str, "blend"))
        s->p_type = VPE_PROCESSOR_BLEND;

    if (!strcmp(s->p_type_str, "tile"))
        s->p_type = VPE_PROCESSOR_TILE;

    if (!strcmp(s->p_type_str, "overlay"))
        s->p_type = VPE_PROCESSOR_OVERLAY;

    if (!strcmp(s->p_type_str, "drawbox"))
        s->p_type = VPE_PROCESSOR_DRAWBOX;

    if (!strcmp(s->p_type_str, "rotate"))
        s->p_type = VPE_PROCESSOR_ROTATE;

    if (!strcmp(s->p_type_str, "csc"))
        s->p_type = VPE_PROCESSOR_CSC;

    if (!strcmp(s->p_type_str, "subsample"))
        s->p_type = VPE_PROCESSOR_SUBSAMPLE;

    /* processor_type of gc620 should be set in a continuous range */
    if (s->p_type <  VPE_PROCESSOR_BLEND  ||
        s->p_type >  VPE_PROCESSOR_ROTATE ||
        s->p_type == VPE_PROCESSOR_DRAWBOX) {
        av_log(ctx, AV_LOG_ERROR, "processor type %d is out of range.\n",
               s->p_type);
        return AVERROR(EINVAL);
    }

    if (s->p_type == VPE_PROCESSOR_DRAWBOX) {
        if (av_parse_color(s->rgb_color, s->color_str, -1, ctx) < 0) {
            av_log(ctx, AV_LOG_ERROR,
                   "Please set correct color value match color_table.\n");
            return AVERROR(EINVAL);
        }

        if (s->nb_inputs != 1) {
            av_log(ctx, AV_LOG_ERROR,
                   "Drawbox only need 1 input stream.\n");
            return AVERROR(EINVAL);
        }

        if (s->thickness > 16) {
            av_log(ctx, AV_LOG_ERROR, "the draw line thickness exceed limitation: %d.\n", s->thickness);
            return AVERROR(EINVAL);
        }
    }

    if (s->p_type == VPE_PROCESSOR_OVERLAY) {
        if (s->nb_inputs != 2) {
            av_log(ctx, AV_LOG_ERROR,
                   "Overlay needs 2 input streams.\n");
            return AVERROR(EINVAL);
        }

        if (s->overlay_x % 2 != 0 || s->overlay_y % 2 != 0) {
            av_log(ctx, AV_LOG_ERROR, "Please use even value for overlay x and y's position.\n");
            return AVERROR(EINVAL);
        }
    }

    if (s->p_type == VPE_PROCESSOR_BLEND) {
        if (s->nb_inputs != 3) {
            av_log(ctx, AV_LOG_ERROR,
                   "Blend need 3 input streams.\n");
            return AVERROR(EINVAL);
        }
    }

    if (s->p_type == VPE_PROCESSOR_ROTATE) {
        if (s->nb_inputs != 1) {
            av_log(ctx, AV_LOG_ERROR,
                   "Rotate need support 1 input stream.\n");
            return AVERROR(EINVAL);
        }

        if (!(s->angle == ROTATE_90_DEGREE || s->angle == ROTATE_180_DEGREE || s->angle == ROTATE_270_DEGREE)) {
            av_log(ctx, AV_LOG_ERROR,
                   "Rotate only support 90, 180, 270 degrees.\n");
            return AVERROR(EINVAL);
        }
    }

    if (s->p_type == VPE_PROCESSOR_CSC) {
        if (s->nb_inputs != 1) {
            av_log(ctx, AV_LOG_ERROR,
                   "CSC only support 1 input stream.\n");
            return AVERROR(EINVAL);
        }

        if (strcmp(s->csc, "rgb2yuv") && strcmp(s->csc, "yuv2rgb")) {
            av_log(ctx, AV_LOG_ERROR,
                   "CSC only support rgb2yuv and yuv2rgb convert.\n");
            return AVERROR(EINVAL);
        }
    }

    if (s->p_type == VPE_PROCESSOR_SUBSAMPLE) {
        if (s->nb_inputs != 1) {
            av_log(ctx, AV_LOG_ERROR,
                   "Subsample only support 1 input stream.\n");
            return AVERROR(EINVAL);
        }
    }

    if (s->p_type == VPE_PROCESSOR_TILE) {
        if (s->x * s->y > 32) {
            av_log(ctx, AV_LOG_ERROR, "tile gride exceed the limitation: %dx%d.\n", s->x, s->y);
            return AVERROR(EINVAL);
        }

        if (s->x * s->y != s->nb_inputs) {
            av_log(ctx, AV_LOG_ERROR, "tile gride not align the input stream number: %dx%d != %d.\n", s->x, s->y, s->nb_inputs);
            return AVERROR(EINVAL);
        }
    }

    av_log(ctx, AV_LOG_DEBUG, "grid x=%d, y=%d.\n", s->x, s->y);

    s->frames = av_calloc(s->nb_inputs, sizeof(*s->frames));
    if (!s->frames)
        return AVERROR(ENOMEM);

    for (i = 0; i < s->nb_inputs; i++) {
        AVFilterPad pad = { 0 };

        pad.type = AVMEDIA_TYPE_VIDEO;
        pad.name = av_asprintf("input%d", i);
        if (!pad.name)
            return AVERROR(ENOMEM);

        if ((ret = ff_append_inpad_free_name(ctx, &pad)) < 0) {
            av_freep(&pad.name);
            return ret;
        }
    }

    return 0;
}

static av_cold int init(AVFilterContext *ctx)
{
    int ret;
    VpeGC620Context *s = ctx->priv;
    set_instance_name(s, ctx->name);

    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GC, s->instance_name);
    ret = vpe_2d_init(ctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();

    return ret;
}

static int process_frame(FFFrameSync *fs)
{
    AVFilterContext *ctx = fs->parent;
    AVFilterLink *outlink = ctx->outputs[0];
    VpeGC620Context *s = fs->opaque;
    AVFrame **in = s->frames;
    AVFrame *out;
    VpiFrm *pic_in, *pic_out, *pic_temp, *pic;
    int i, ret;

    for (i = 0; i < s->nb_inputs; i++) {
        if ((ret = ff_framesync_get_frame(&s->fs, i, &in[i], 0)) < 0)
            return ret;

        /* set vpi input frames */
        pic_temp = (VpiFrm *)(in[i]->data[0]);
        if (!pic_temp)
            return AVERROR_INVALIDDATA;

        if (i == 0) {
            pic_in = pic_temp;
            pic    = pic_in;
        } else {
            pic->next = pic_temp;
            pic       = pic->next;
        }
    }

    /* need to get new video buffer */
    out = ff_get_video_buffer(outlink, outlink->w, outlink->h);
    if (!out)
        return AVERROR(ENOMEM);

    out->pts = av_rescale_q(s->fs.pts, s->fs.time_base, outlink->time_base);
    out->sample_aspect_ratio = outlink->sample_aspect_ratio;

    pic_out = (VpiFrm *)(out->data[0]);
    if (!pic_out)
        return AVERROR_INVALIDDATA;

    /* set the channle index for next chain */
    if (!(out->buf[1]) && s->out_chan_index) {
        int *index;

        out->buf[1] = av_buffer_alloc(sizeof(int));
        index = out->buf[1]->data;
        *index = s->out_chan_index;
    }

    ret = vpi_processor_process_frame(s->vpe_gc620_handle, pic_in, pic_out);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "failed to process frames in@%p, out@%p, ret=%s\n",
               pic_in, pic_out, vpi_error_str(ret));
        return AVERROR_EXTERNAL;
    }

    return ff_filter_frame(outlink, out);
}

static int vpe_2d_config_output(AVFilterLink *outlink)
{
    AVFilterContext *ctx = outlink->src;
    VpeGC620Context *s = ctx->priv;
    AVRational frame_rate = ctx->inputs[0]->frame_rate;
    AVRational sar = ctx->inputs[0]->sample_aspect_ratio;
    int height = ctx->inputs[0]->h;
    int width = ctx->inputs[0]->w;
    FFFrameSyncIn *in;
    int i, ret, ch_index = 0;
    VpiProcInitParams *params = &s->vpe_gc620_params;
    AVHWFramesContext* hwframe;
    VpiChannelInfo in_ch_info[VPE_CHANNEL_NUM] = { 0 };

    s->desc = av_pix_fmt_desc_get(outlink->format);
    if (!s->desc)
        return AVERROR_BUG;

    if (s->p_type == VPE_PROCESSOR_ROTATE) {
        if (s->angle == ROTATE_90_DEGREE || s->angle == ROTATE_270_DEGREE) {
          height = ctx->inputs[0]->w;
          width  = ctx->inputs[0]->h;
        }
    }
    if (s->p_type == VPE_PROCESSOR_TILE) {
        height = s->x * height;
        width  = s->y * width;
    }
    if (s->width > 0 && s->height > 0){
        height = s->height;
        width  = s->width;
    }
    /* limit to 8k */
    width  = width  > OUT_MAX ? OUT_MAX : width;
    height = height > OUT_MAX ? OUT_MAX : height;
    outlink->w          = width;
    outlink->h          = height;
    outlink->frame_rate = frame_rate;
    outlink->sample_aspect_ratio = sar;

    if ((ret = ff_framesync_init(&s->fs, ctx, s->nb_inputs)) < 0)
        return ret;

    in = s->fs.in;
    s->fs.opaque = s;
    s->fs.on_event = process_frame;

    for (i = 0; i < s->nb_inputs; i++) {
        AVFilterLink *inlink = ctx->inputs[i];

        in[i].time_base = inlink->time_base;
        in[i].sync   = 1;
        in[i].before = EXT_STOP;
        in[i].after  = s->shortest ? EXT_STOP : EXT_INFINITY;
    }

    if(ff_framesync_configure(&s->fs) < 0) {
        av_log(ctx, AV_LOG_ERROR, "failed to configure ff_framesync.\n");
        return AVERROR(EINVAL);
    }
    outlink->time_base = s->fs.time_base;

    /* init vpe processor handle */
    ret = vpe_init_hwctx(ctx);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "Vpe init hwctx failed.\n");
        return AVERROR_EXTERNAL;
    }

    /* change format for csc/subsample filter */
    outlink->hw_frames_ctx = av_hwframe_ctx_alloc(s->hwdevice);
    if(!outlink->hw_frames_ctx) {
        av_log(ctx, AV_LOG_ERROR, "hw_frames_ctx alloc failed\n");
        return AVERROR(ENOMEM);
    }

    hwframe = (AVHWFramesContext*) outlink->hw_frames_ctx->data;
    AVHWFramesContext* fmctx = (AVHWFramesContext*) (ctx->inputs[0]->hw_frames_ctx->data);
    AVVpeFramesContext* inlink_frame_hwctx = (AVVpeFramesContext*)fmctx->hwctx;

    hwframe->width           = width;
    hwframe->height          = height;
    hwframe->format          = AV_PIX_FMT_VPE;
    hwframe->sw_format       = fmctx->sw_format;

    if (s->p_type == VPE_PROCESSOR_CSC) {
        hwframe->sw_format = isRGB(fmctx->sw_format) ? AV_PIX_FMT_YUV420P : AV_PIX_FMT_BGRA; /* todo: get from param */
    }
    else if (s->p_type == VPE_PROCESSOR_SUBSAMPLE) {
        if (fmctx->sw_format == AV_PIX_FMT_YUV422P || fmctx->sw_format == AV_PIX_FMT_YUV420P)
            hwframe->sw_format = AV_PIX_FMT_YUV420P;
        else {
            av_log(ctx, AV_LOG_ERROR, "format %s not support for subsample\n", av_get_pix_fmt_name(fmctx->sw_format));
            return AVERROR(EINVAL);
        }
    }

    if ((ret = av_hwframe_ctx_init(outlink->hw_frames_ctx)) < 0) {
        av_log(ctx, AV_LOG_ERROR, "av_hwframe_ctx_init for output failed\n");
        return AVERROR(EINVAL);
    }

    /* get input channle info */
    ret = vpi_memory_get_channel_info(inlink_frame_hwctx->vpe_mem_ctx, in_ch_info);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpi_memory_get_channel_info failed.\n");
        return AVERROR(EINVAL);
    }

    ret = vpi_memory_get_map_info(inlink_frame_hwctx->vpe_mem_ctx, 0, &ch_index);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpi_memory_get_map_info failed.\n");
        return AVERROR(EINVAL);
    }

    AVVpeFramesContext* outlink_frame_hwctx              = (AVVpeFramesContext*)hwframe->hwctx;
    outlink_frame_hwctx->proc_info->ch_info[0]           = in_ch_info[ch_index];
    outlink_frame_hwctx->proc_info->ch_info[0].enabled   = 1;
    outlink_frame_hwctx->proc_info->ch_info[0].format    = VPI_FORMAT_VPE;
    outlink_frame_hwctx->proc_info->ch_info[0].sw_format = get_vpifmt_from_avpixfmt(hwframe->sw_format);

    //set map info into
    ret = vpi_memory_set_map_info(outlink_frame_hwctx->vpe_mem_ctx, 0, 0);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpi_memory_set_map_info failed, ret=%d\n", ret);
        return AVERROR_EXTERNAL;
    }

    //set chan info
    ret = vpi_memory_set_channel_info(outlink_frame_hwctx->vpe_mem_ctx, outlink_frame_hwctx->proc_info->ch_info, VPE_HW_DRAWBOX);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpi_memory_set_single_channel_info failed, ret=%d\n", ret);
        return AVERROR_EXTERNAL;
    }

    s->hw_frame_ctx = av_buffer_ref(outlink->hw_frames_ctx);
    s->out_chan_index = 0;

    if (!outlink->hw_frames_ctx) {
        av_log(ctx, AV_LOG_ERROR, "outlink hw frame context get failed.\n");
        return AVERROR(ENOMEM);
    }

    /* get params */
    if (ctx->inputs[0]->format == AV_PIX_FMT_VPE) {
        fmctx = (AVHWFramesContext *)(ctx->inputs[0]->hw_frames_ctx->data);
        params->input_sw_format = av_get_pix_fmt_name(fmctx->sw_format);
        params->input_format    = av_get_pix_fmt_name(ctx->inputs[0]->format);
        av_log(ctx, AV_LOG_DEBUG,
               "%s(%d), fmctx->sw_format=%d, params->input_sw_format=%s.\n",
               __func__, __LINE__, fmctx->sw_format, params->input_sw_format);
    } else {
        av_log(ctx, AV_LOG_ERROR, "%s: Format: %s not supported.\n", __func__,
               av_get_pix_fmt_name(ctx->inputs[0]->format));
        return AVERROR(EINVAL);
    }

    if (s->p_type == VPE_PROCESSOR_OVERLAY) {
        if (s->overlay_x > width || s->overlay_y > height) {
            av_log(ctx, AV_LOG_ERROR,
                   "Please set overlay x and y's position less than input resolution boundary.\n");
            return AVERROR(EINVAL);
        }
    }

    params->processor_type = s->p_type;
    params->core_id        = s->core_id;
    params->grid_x         = s->x;
    params->grid_y         = s->y;
    params->nb_frames      = s->nb_inputs;
    params->width          = width;
    params->height         = height;
    params->x_position     = s->overlay_x;
    params->y_position     = s->overlay_y;
    params->box_thickness  = s->thickness;
    params->box_color = ((((s->rgb_color[A]) & 0xFF) << 24) | (((s->rgb_color[R])&0xFF) << 16) |
	                    (((s->rgb_color[G])&0xFF) << 8) | ((s->rgb_color[B])&0xFF));
    params->rotate_angle   = s->angle;
    params->csc            = strcmp(s->csc, "yuv2rgb") ? 0 : 1;

    av_log(ctx, AV_LOG_DEBUG,
        "params: grid_x=%d, grid_y=%d, nb_frames=%d, width=%d, height=%d.\n",
        params->grid_x, params->grid_y, params->nb_frames, params->width,
        params->height);

    ret = vpi_processor_init(params, &s->vpe_gc620_handle);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "%s(%d): processor init failed: %s!\n",
               __func__, __LINE__, vpi_error_str(ret));
        return AVERROR_EXTERNAL;
    }

    ret = vpi_processor_open(s->vpe_gc620_handle);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "%s(%d): processor open failed: %s.\n",
               __func__, __LINE__, vpi_error_str(ret));
        return AVERROR_EXTERNAL;
    }
    return ret;
}

static int config_output(AVFilterLink *outlink)
{
    int ret;
    AVFilterContext *ctx = outlink->src;
    VpeGC620Context *s = ctx->priv;
    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GC, s->instance_name);
    ret = vpe_2d_config_output(outlink);
    LOG_AMA_RESTORE_INSTANCE_INFO();

    return ret;
}

static av_cold void vpe_2d_uninit(AVFilterContext *ctx)
{
    VpeGC620Context *s = ctx->priv;

    ff_framesync_uninit(&s->fs);
    av_freep(&s->frames);

    av_buffer_unref(&s->hwdevice);
    av_buffer_unref(&s->hw_frame_ctx);

    if (s->vpe_gc620_handle)
        vpi_processor_close(s->vpe_gc620_handle);
}

static av_cold void uninit(AVFilterContext *ctx)
{
    VpeGC620Context *s = ctx->priv;

    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GC, s->instance_name);
    vpe_2d_uninit(ctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();
}

static int vpe_2d_activate(AVFilterContext *ctx)
{
    VpeGC620Context *s = ctx->priv;
    return ff_framesync_activate(&s->fs);
}

static int activate(AVFilterContext *ctx)
{
    int ret;
    VpeGC620Context *s = ctx->priv;

    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GC, s->instance_name);
    ret = vpe_2d_activate(ctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();

    return ret;
}

#define OFFSET(x) offsetof(VpeGC620Context, x)
#define FLAGS AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_FILTERING_PARAM

static const AVFilterPad gc620_outputs[] = {
    {
        .name          = "default",
        .type          = AVMEDIA_TYPE_VIDEO,
        .config_props  = config_output,
    },
};

static const AVOption gpu_2d_ama_options[] = {
    { "inputs", "set number of inputs", OFFSET(nb_inputs), AV_OPT_TYPE_INT, {.i64=2}, 1, 32, .flags = FLAGS },
    { "processor",  "set processor type", OFFSET(p_type_str), AV_OPT_TYPE_STRING, {.str = "none"}, .flags = FLAGS, .unit="processor" },
        { "blend",     "blend multiple sources together using an alpha blending mask", 0, AV_OPT_TYPE_CONST, {.str = "blend"},     .flags = FLAGS, .unit="processor" },
        { "tile",      "map input streams to specified layout",                        0, AV_OPT_TYPE_CONST, {.str = "tile"},      .flags = FLAGS, .unit="processor" },
        { "overlay",   "overlay 1 picture over another",                               0, AV_OPT_TYPE_CONST, {.str = "overlay"},   .flags = FLAGS, .unit="processor" },
        { "rotate",    "rotate input by specified angle",                              0, AV_OPT_TYPE_CONST, {.str = "rotate"},    .flags = FLAGS, .unit="processor" },
        { "csc",       "color space conversion from rgb2yuv or yuv2rgb",               0, AV_OPT_TYPE_CONST, {.str = "csc"},       .flags = FLAGS, .unit="processor" },
        { "subsample", "chroma subsample from 422 to 420",                             0, AV_OPT_TYPE_CONST, {.str = "subsample"}, .flags = FLAGS, .unit="processor" },
    { "layout", "set grid size (for tiling processor) [Max 32 inputs]", OFFSET(x), AV_OPT_TYPE_IMAGE_SIZE, {.str="2x2"}, 0, 0, .flags = FLAGS},
    { "core_id", "set core to run", OFFSET(core_id), AV_OPT_TYPE_INT, {.i64=0}, 0, 1, .flags = FLAGS},
    { "shortest", "force termination when the shortest input terminates", OFFSET(shortest), AV_OPT_TYPE_BOOL, {.i64=0}, 0, 1, .flags = FLAGS },
    { "x", "Overlay x_position", OFFSET(overlay_x), AV_OPT_TYPE_INT, { .i64=0 }, 0, INT_MAX, .flags = FLAGS },
    { "y", "Overlay y_position", OFFSET(overlay_y), AV_OPT_TYPE_INT, { .i64=0 }, 0, INT_MAX, .flags = FLAGS },
    { "rotation", "set the angle (for rotate processor)", OFFSET(angle), AV_OPT_TYPE_INT, { .i64=0 }, 0, 270, .flags = FLAGS, .unit="rotation" },
        { "90",  "rotate by  90 degrees", 0, AV_OPT_TYPE_CONST, { .i64=90 },  0, 0, FLAGS, .unit = "rotation"},
        { "180", "rotate by 180 degrees", 0, AV_OPT_TYPE_CONST, { .i64=180 }, 0, 0, FLAGS, .unit = "rotation"},
        { "270", "rotate by 270 degrees", 0, AV_OPT_TYPE_CONST, { .i64=270 }, 0, 0, FLAGS, .unit = "rotation"},
    { "csc", "set the color space conversion required (rgb2yuv or yuv2rgb for csc processor)", OFFSET(csc), AV_OPT_TYPE_STRING, { .str="none" }, .flags = FLAGS },
    { "w", "set the output resolution width  [optional]", OFFSET(width), AV_OPT_TYPE_INT, { .i64=0 }, 0, OUT_MAX, .flags = FLAGS },
    { "h", "set the output resolution height [optional]", OFFSET(height), AV_OPT_TYPE_INT, { .i64=0 }, 0, OUT_MAX, .flags = FLAGS },
    { NULL },
};

AVFILTER_DEFINE_CLASS(gpu_2d_ama);

const AVFilter ff_vf_gc620_vpe = {
    .name          = "2d_ama",
    .description   = NULL_IF_CONFIG_SMALL("2D Engine: blend tile overlay rotate csc subsample"),
    .priv_size     = sizeof(VpeGC620Context),
    .priv_class    = &gpu_2d_ama_class,
    FILTER_OUTPUTS(gc620_outputs),
    FILTER_QUERY_FUNC(query_formats),
    .init          = init,
    .uninit        = uninit,
    .activate      = activate,
    .flags         = AVFILTER_FLAG_DYNAMIC_INPUTS | AVFILTER_FLAG_SLICE_THREADS,
    .flags_internal = FF_FILTER_FLAG_HWFRAME_AWARE,
};
