/*
 * AMA sideband data write filter
 * Copyright (C) 2023 VeriSilicon Holdings Co., Ltd.
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file
 * AMA sideband data write filter
 */
#include "vf_sd_common.h"

#define FLAGS AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_FILTERING_PARAM

static av_cold int init(AVFilterContext *ctx)
{
    AmaSdContext *sb = NULL;

    sb = ctx->priv;
    if(!sb->sb_file || sb->sb_file == ""){
        av_log(NULL, AV_LOG_ERROR, "Please set a valid file path for sdwrite\n");
        return AVERROR(EINVAL);
    }
    sb->sb_fp = fopen(sb->sb_file, "wb");
    if(!sb->sb_fp){
        av_log(NULL, AV_LOG_ERROR, "Failed to open file %s\n", sb->sb_file);
        return AVERROR(EINVAL);
    }
    return 0;
}

static av_cold void uninit(AVFilterContext *ctx)
{
    AmaSdContext *sb = ctx->priv;

    if(sb->sb_fp) {
        fclose(sb->sb_fp);
    }
}

static int filter_frame(AVFilterLink *inlink, AVFrame *in)
{
    AVFilterLink *outlink = inlink->dst->outputs[0];
    AVFrame *out = ff_get_video_buffer(outlink, in->width, in->height);
    AmaSdContext *sb = NULL;
    int ret , sb_type;
    AVFrameSideData *sb_info = NULL;
    AVDictionaryEntry *t = NULL;
    uint32_t frameId;

    sb = inlink->dst->priv;

    if (!out) {
        ret = AVERROR(ENOMEM);
        goto fail;
    }
    ret = av_frame_copy_props(out, in);
    if (ret < 0)
        goto fail;

    ret = av_frame_copy(out, in);
    if (ret < 0)
        goto fail;

    for (sb_type = 0; sb_type < NUM_AVFRAME_SIDEDATA_TYPE; sb_type++){
        sb_info = av_frame_get_side_data(out, sb_type);
        if (!sb_info) {
            continue;
        }
        frameId = 0;
        t = av_dict_get(sb_info->metadata, "frameId", NULL, 0);
        if (t)
            frameId = atoi(t->value);
        sb->com_header.frameId = frameId;
        sb->com_header.type = sb_info->type;
        sb->com_header.data_size = sb_info->size;

        av_log(NULL, AV_LOG_DEBUG, "sd_write: frameId %d, type %d, size %d\n", frameId, sb_info->type, sb_info->size);
        switch (sb_info->type)
        {
            case AV_FRAME_DATA_MASTERING_DISPLAY_METADATA:
            case AV_FRAME_DATA_CONTENT_LIGHT_LEVEL:
            case AV_FRAME_DATA_DYNAMIC_HDR_PLUS:
            case AV_FRAME_DATA_VUI_PARAMS: {
                fwrite(&sb->com_header, COMMON_HEAD_SIZE, 1, sb->sb_fp);
                fwrite(sb_info->data, sizeof(uint8_t), sb_info->size, sb->sb_fp);
                break;
            }
            case AV_FRAME_DATA_RAW_T35: {
                AVRawT35Data *raw_data = (AVRawT35Data *)sb_info->data;
                sb->com_header.data_size = raw_data->payload_byte_length;
                // write overall file
                fwrite(&sb->com_header, COMMON_HEAD_SIZE, 1, sb->sb_fp);
                fwrite(&raw_data->payload_byte_length, sizeof(uint32_t), 1, sb->sb_fp);
                fwrite(raw_data->payload_byte_data, sizeof(uint8_t), raw_data->payload_byte_length, sb->sb_fp);
                break;
            }
            case AV_FRAME_DATA_SB_ML_ROI:
            case AV_FRAME_DATA_SB_ML_CC:
            case AV_FRAME_DATA_SB_ML_CXC:
            case AV_FRAME_DATA_SB_ML_LA_STAT:
            case AV_FRAME_DATA_SB_ML_BOUNDING_BOX:
            case AV_FRAME_DATA_SB_ML_CLASSIFICATION:
            case AV_FRAME_DATA_SB_ML_YDATA:
            case AV_FRAME_DATA_SB_ML_TENSOR: {
                uint32_t width = 0;
                uint32_t height = 0;
                uint32_t channel = 0;
                uint32_t batch = 0;
                uint32_t layout = VPE_LAYOUT_NCHW;
                t = av_dict_get(sb_info->metadata, "width", NULL, 0);
                if (t)
                    width = atoi(t->value);
                t = av_dict_get(sb_info->metadata, "height", NULL, 0);
                if (t)
                    height = atoi(t->value);
                t = av_dict_get(sb_info->metadata, "channel", NULL, 0);
                if (t)
                    channel = atoi(t->value);
                t = av_dict_get(sb_info->metadata, "batch", NULL, 0);
                if (t)
                    batch = atoi(t->value);
                t = av_dict_get(sb_info->metadata, "layout", NULL, 0);
                if (t)
                    layout = atoi(t->value);
                sb->ml_header.width   = width;
                sb->ml_header.height  = height;
                sb->ml_header.channel = channel;
                sb->ml_header.batch   = batch;
                sb->ml_header.layout  = layout;
                fwrite(&sb->com_header, COMMON_HEAD_SIZE, 1, sb->sb_fp);
                fwrite(&sb->ml_header, ML_HEAD_SIZE, 1, sb->sb_fp);
                fwrite(sb_info->data, sizeof(uint8_t), sb_info->size, sb->sb_fp);
                break;
            }
            default:
                break;
        }
    }
    av_frame_free(&in);
    return ff_filter_frame(outlink, out);
fail:
    av_frame_free(&in);
    av_frame_free(&out);
    return ret;
}

static const AVOption sd_write_options[] = {
    { "file",   "sidedata file path", OFFSET(sb_file), AV_OPT_TYPE_STRING, { .str = NULL },  .flags = FLAGS },
    { NULL }
};

AVFILTER_DEFINE_CLASS(sd_write);

static const AVFilterPad avfilter_vf_sd_write_inputs[] = {
    {
        .name         = "default",
        .type         = AVMEDIA_TYPE_VIDEO,
        .filter_frame = filter_frame,
    },
};

static const AVFilterPad avfilter_vf_sd_write_outputs[] = {
    {
        .name = "default",
        .type = AVMEDIA_TYPE_VIDEO,
    },
};

const AVFilter ff_vf_sd_write = {
    .name        = "sdwrite_ama",
    .description = NULL_IF_CONFIG_SMALL("Write sidedata data to file."),
    FILTER_INPUTS(avfilter_vf_sd_write_inputs),
    FILTER_OUTPUTS(avfilter_vf_sd_write_outputs),
    .priv_size   = sizeof(AmaSdContext),
    .priv_class  = &sd_write_class,
    .init        = init,
    .uninit      = uninit,
    .nb_outputs  = 1,
};