/*
 * VeriSilicon VPE Split Filter
 * Copyright (C) 2020 VeriSilicon Holdings Co., Ltd.
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "vpi_types.h"
#include "vpi_api.h"

#include "filters.h"
#include "libavutil/opt.h"
#include "libavutil/buffer.h"
#include "libavutil/buffer_internal.h"
#include "libavutil/avstring.h"
#include "libavutil/hwcontext.h"
#include "libavutil/hwcontext_vpe.h"

typedef struct VpeSplitContext {
    const AVClass *class;
    /*setting output channel number*/
    int nb_outputs;
    /*valid output channel number*/
    int valid_outputs;
    /*need new output vpiframe flag*/
    int new_out_flag;
    VpiProcHandle vpe_split_ctx;
} VpeSplitContext;

#define VPE_SPLIT_MAX   6

static int vpe_split_out_config_props(AVFilterLink *outlink)
{
    AVFilterContext *src = outlink->src;
    int out_index;
    AVHWFramesContext *hwframe_ctx;
    AVVpeFramesContext *vpeframe_ctx;
    int ret;
    int out_ch;

    if (!src->inputs[0]->hw_frames_ctx)
        return 0;

    outlink->hw_frames_ctx  = av_buffer_ref(src->inputs[0]->hw_frames_ctx);
    hwframe_ctx             = (AVHWFramesContext *)outlink->hw_frames_ctx->data;
    vpeframe_ctx            = (AVVpeFramesContext *)hwframe_ctx->hwctx;

    for (out_index = 0; out_index < src->nb_outputs; out_index++) {
        if (outlink == src->outputs[out_index])
            break;
    }
    if (out_index == src->nb_outputs) {
        av_log(src, AV_LOG_ERROR, "can't find output\n");
        return AVERROR_INVALIDDATA;
    }
    av_log(src, AV_LOG_DEBUG, "out_index %d\n", out_index);

    ret = vpi_memory_get_channel_info(vpeframe_ctx->vpe_mem_ctx, &vpeframe_ctx->proc_info->ch_info[0]);
    if (ret < 0) {
        av_log(src, AV_LOG_ERROR, "vpi_memory_get_info failed\n");
        return AVERROR_EXTERNAL;
    }
    ret = vpi_memory_get_map_info(vpeframe_ctx->vpe_mem_ctx, out_index, &out_ch);
    if (ret < 0) {
        av_log(src, AV_LOG_ERROR, "vpi_memory_get_map_info failed\n");
        return AVERROR_EXTERNAL;
    }
    if (vpeframe_ctx->proc_info->ch_info[out_ch].enabled) {
        outlink->w = vpeframe_ctx->proc_info->ch_info[out_ch].width;
        outlink->h = vpeframe_ctx->proc_info->ch_info[out_ch].height;
        av_log(src, AV_LOG_DEBUG, "out_ch %d, w %d, h %d\n", out_ch, outlink->w, outlink->h);
    } else {
        av_log(src, AV_LOG_ERROR, "%d channel not enabled\n", out_ch);
        return AVERROR_EXTERNAL;
    }

    return 0;
}

static av_cold int vpe_split_init(AVFilterContext *ctx)
{
    VpeSplitContext *s = ctx->priv;
    int i, ret;

    for (i = 0; i < s->nb_outputs; i++) {
        AVFilterPad pad = { 0 };

        pad.type = AVMEDIA_TYPE_VIDEO;
        pad.name = av_asprintf("output%d", i);
        if (!pad.name)
            return AVERROR(ENOMEM);
        pad.config_props = vpe_split_out_config_props;

        if ((ret = ff_insert_outpad(ctx, i, &pad)) < 0) {
            av_freep(&pad.name);
            return ret;
        }
    }

    return 0;
}

static av_cold void vpe_split_uninit(AVFilterContext *ctx)
{
    VpeSplitContext *s = (VpeSplitContext *)ctx->priv;
    int i;

    for (i = 0; i < ctx->nb_outputs; i++) {
        av_freep(&ctx->output_pads[i].name);
    }

    vpi_processor_close(s->vpe_split_ctx);
}

static int vpe_split_config_props(AVFilterLink *inlink)
{
    AVHWDeviceContext *device_ctx;
    AVVpeDeviceContext *device_hwctx;
    AVHWFramesContext *hwframe_ctx;
    AVVpeFramesContext * frame_hwctx;
    AVFilterContext *dst = inlink->dst;
    VpeSplitContext *s = dst->priv;
    VpiProcInitParams *params;
    int ret;

    hwframe_ctx = (AVHWFramesContext *)inlink->hw_frames_ctx->data;
    if (!hwframe_ctx) {
        av_log(dst, AV_LOG_ERROR, "hwframe_ctx is nil\n");
        return AVERROR_UNKNOWN;
    }

    frame_hwctx = hwframe_ctx->hwctx;
    if (!frame_hwctx) {
        av_log(dst, AV_LOG_ERROR, "frame_hwctx is nil\n");
        return AVERROR_UNKNOWN;
    }

    device_ctx = hwframe_ctx->device_ctx;
    device_hwctx = device_ctx->hwctx;

    params = (VpiProcInitParams *)vpi_memory_get_struct(VPE_PROC_INIT_PARAM);
    if (!params) {
        return AVERROR(ENOMEM);
    }
    vpi_device_set_ctx(params, device_hwctx->vpe_dev_ctx, VPE_HW_SPLIT);
    params->mem_ctx = frame_hwctx->vpe_mem_ctx;
    params->processor_type = VPE_PROCESSOR_SPLIT;
    params->output_num = s->nb_outputs;
    ret = vpi_processor_init(params, &s->vpe_split_ctx);
    if (ret < 0) {
        av_log(dst, AV_LOG_ERROR, "VPE processor split init error: %s\n", vpi_error_str(ret));
        vpi_memory_free_struct(params);
        return AVERROR_EXTERNAL;
    }

    vpi_memory_free_struct(params);

    return 0;
}

static void vpe_split_free_buffer(void *opaque, uint8_t *data)
{
    VpiFrm *vpi_frame = (VpiFrm *)data;

    vpi_frame_release(vpi_frame);
    av_buffer_pool_release_buffer(opaque, data);
}

static int split_vpe_get_frame(AVFilterLink *outlink, AVFrame *in, AVFrame **out_avfrm)
{
    AVFrame *out = NULL;
    VpiFrm *out_pic = NULL;
    VpiFrm *in_pic = NULL;
    int ret = 0;

    if (!outlink || !in)
        return AVERROR(EINVAL);

    out = ff_get_video_buffer(outlink, in->width, in->height);
    if (!out) {
        av_log(NULL, AV_LOG_ERROR, "failed to allocate frame.");
        return AVERROR(ENOMEM);
    }

    ret = av_frame_copy_props(out, in);
    if (ret < 0){
        av_log(NULL, AV_LOG_ERROR, "av_frame_copy_props failed\n");
        return AVERROR_EXTERNAL;
    }

    out_pic = (VpiFrm *)(out->data[0]);
    if (!out_pic){
        return AVERROR_INVALIDDATA;
    }
    in_pic = (VpiFrm *)(in->data[0]);

    //out_pic shall be same with input pic
    out_pic->vpi_picinfo_buf      = in_pic->vpi_picinfo_buf;
    out_pic->width                = in_pic->width;
    out_pic->height               = in_pic->height;
    out_pic->format               = in_pic->format;
    out_pic->ch_index             = in_pic->ch_index;
    out_pic->ch_info[out_pic->ch_index]    = in_pic->ch_info[in_pic->ch_index];
    vpi_frame_ref(out_pic);

    //update sidedata
    vpifrm_sidedata_copy_reference(in_pic, out_pic);

    *out_avfrm = out;

    return 0;
}

static int vpe_split_filter_frame(AVFilterLink *inlink, AVFrame *frame)
{
    AVFilterContext *ctx = inlink->dst;
    VpeSplitContext *s = ctx->priv;
    AVHWFramesContext *hwframe_ctx;
    VpiFrm *frm;
    int i, ret = AVERROR_EOF;

    hwframe_ctx = (AVHWFramesContext *)inlink->hw_frames_ctx->data;
    if (!hwframe_ctx) {
        av_log(dst, AV_LOG_ERROR, "hwframe_ctx is nil\n");
        return AVERROR_UNKNOWN;
    }

    for (i = 0; i < ctx->nb_outputs; i++) {
        AVFrame *buf_out;
        int *index;

        if (ff_outlink_get_status(ctx->outputs[i]))
            continue;

        /* in case split need new vpiframe */
        if (s->new_out_flag) {
            ret = split_vpe_get_frame(ctx->outputs[i], frame, &buf_out);
            if (ret < 0) {
                av_log(ctx, AV_LOG_ERROR, "split frame failed(%s).\n", vpi_error_str(ret));
                return AVERROR_EXTERNAL;
            }
        } else {
            buf_out = av_frame_clone(frame);
            if (!buf_out) {
                ret = AVERROR(ENOMEM);
                break;
            }
            buf_out->buf[1] = av_buffer_alloc(sizeof(int));
            index = buf_out->buf[1]->data;
            *index = i;
            if (i) {
                if (vpi_processor_process_frame(s->vpe_split_ctx,
                                        (VpiFrm *)frame->buf[0]->data,
                                        (VpiFrm *)buf_out->buf[0]->data)) {
                    av_log(ctx, AV_LOG_ERROR, "vpi_processor_process_frame failed, error=%s\n", vpi_error_str(ret));
                    return AVERROR_EXTERNAL;
                }
            }
        }

        ret = ff_filter_frame(ctx->outputs[i], buf_out);
        if (ret < 0)
            break;
    }
    av_frame_free(&frame);
    return ret;
}

#define OFFSET(x) offsetof(VpeSplitContext, x)
#define FLAGS (AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_FILTERING_PARAM)
static const AVOption vpe_split_options[] = { { "outputs",
                                                "set number of outputs",
                                                OFFSET(nb_outputs),
                                                AV_OPT_TYPE_INT,
                                                { .i64 = 2 },
                                                1,
                                                VPE_SPLIT_MAX,
                                                FLAGS },
                                              { "new_out_flag",
                                                "set output vpiframe as a new vpiframe",
                                                OFFSET(new_out_flag),
                                                AV_OPT_TYPE_INT,
                                                { .i64 = 1 },
                                                0,
                                                1,
                                                FLAGS },
                                              { NULL }
                                            };

AVFILTER_DEFINE_CLASS(vpe_split);

static const AVFilterPad avfilter_vpe_split_inputs[] =
    { {
          .name         = "default",
          .type         = AVMEDIA_TYPE_VIDEO,
          .config_props = vpe_split_config_props,
          .filter_frame = vpe_split_filter_frame,
      },
      { NULL } };


AVFilter ff_vf_split_vpe = {
    .name           = "split_vpe",
    .description    = NULL_IF_CONFIG_SMALL("Pass on the input to N video vpe outputs."),
    .priv_size      = sizeof(VpeSplitContext),
    .priv_class     = &vpe_split_class,
    .init           = vpe_split_init,
    .uninit         = vpe_split_uninit,
    .inputs         = avfilter_vpe_split_inputs,
    .outputs        = NULL,
    .flags          = AVFILTER_FLAG_DYNAMIC_OUTPUTS,
    .flags_internal = FF_FILTER_FLAG_HWFRAME_AWARE,
};
